import importlib
import importlib.util
import os
import sys
from importlib.machinery import ModuleSpec
from typing import Any, Callable, Optional

from ml_collections import FieldReference


def get_class_path(cls: Callable) -> str:
    """
    Returns the full path of a class as a string.

    Copied from https://github.com/Farama-Foundation/HighwayEnv
    """
    return cls.__module__ + "." + cls.__qualname__


def class_from_path(path: str) -> Callable:
    """
    Returns a class object from its full path.
    The path should be in the format 'module.ClassName'.

    Copied from https://github.com/Farama-Foundation/HighwayEnv
    """
    module_name, class_name = path.rsplit(".", 1)
    class_object = getattr(importlib.import_module(module_name), class_name)
    return class_object


def instantiate_object_from_config(config: dict) -> Callable:
    """
    Instantiates a class from a configuration dictionary.
    The dictionary must contain a 'type' key with the full class path.
    """
    assert "type" in config, "Config must contain 'type' key"
    class_path = config["type"]
    params = {k: v for k, v in config.items() if not k == "type"}
    class_object = class_from_path(class_path)
    return class_object(**params)


def load_config(*, config_path: str, overrides: Optional[list[str]] = None) -> dict:
    """
    Load a Python configuration file and extract its variables into a dictionary.
    Variables that start with an underscore, callable objects, and modules are filtered out.
    """
    config_path = os.path.abspath(config_path)
    spec = importlib.util.spec_from_file_location("user_config", config_path)
    assert isinstance(spec, ModuleSpec), "Failed to load module spec"

    config = importlib.util.module_from_spec(spec)
    sys.modules["user_config"] = config
    spec.loader.exec_module(config)

    config = vars(config)
    _update_config_with_overrides(config=config, overrides=overrides)

    # Extract variables: filter out built-ins and modules
    config_dict = _resolve_config_dict(config)
    return config_dict


def _resolve_config_dict(config: dict) -> dict:
    def _resolve_value(value: Any) -> Any:
        if isinstance(value, dict):
            return _resolve_config_dict(value)
        elif isinstance(value, FieldReference):
            return value.get()
        return value

    config_dict = {
        key: _resolve_value(value)
        for key, value in config.items()
        if not key.startswith("_")
        and not callable(value)
        and not isinstance(value, type(sys))
    }
    return config_dict


def _update_config_with_overrides(
    *, config: dict, overrides: Optional[list[str]]
) -> dict:
    """
    Update the configuration dictionary with command line overrides.
    Each override should be in the format 'key=value'.
    e.g. "--override port=9090 --override config.batch_size=32"
    """
    if overrides:
        for item in overrides:
            key, value = item.split("=", 1)
            _set_nested(config, key, value)
    return config


def _set_nested(config, var_path, value):
    """
    Set a nested value in config given a dotted path (e.g., my_dict.name).
    """
    keys = var_path.split(".")
    obj = config
    for key in keys[:-1]:
        obj = getattr(obj, key) if hasattr(obj, key) else obj[key]

    # Convert value to appropriate type
    if value.lower() == "true":
        value = True
    elif value.lower() == "false":
        value = False
    else:
        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                pass

    # handle FieldReference overrides
    if isinstance(obj[keys[-1]], FieldReference):
        obj[keys[-1]].set(value)
    else:
        obj[keys[-1]] = value



"""
Config file for testing
"""
from base_config import class_config_2 as class_config_2_

from konfigurator import FieldReference

_work_dir = FieldReference("tests/work_dir")
experiment_dir = _work_dir + "/experiment"

class_config_1 = dict(
    type="tests.test_utils.MyTestClass",
    name="test_instance_1",
)
class_config_2 = class_config_2_



import json
import os
import subprocess
import tempfile

from konfigurator import load_config

this_dir = os.path.dirname(os.path.abspath(__file__))
config_path = os.path.join(this_dir, "config.py")
script_path = os.path.join(this_dir, "script.py")


def test_load_config():
    config = load_config(config_path=config_path)
    assert isinstance(config, dict), "Config should be a dictionary"
    assert "experiment_dir" in config, "Config should contain 'experiment_dir'"
    assert "class_config_1" in config, "Config should contain 'class_config_1'"
    assert "class_config_2" in config, "Config should contain 'class_config_2'"
    assert "_test_class_obj" not in config, "Config should contain '_test_class_obj'"
    assert (
        config["experiment_dir"] == "tests/work_dir/experiment"
    ), "Experiment dir should be set correctly"


def test_load_config_with_overrides():
    config = load_config(
        config_path=config_path,
        overrides=[
            "_work_dir=affe",
            "class_config_1.name=overridden_name",
            "class_config_2.name=overridden_name2",
        ],
    )
    assert (
        config["class_config_1"]["name"] == "overridden_name"
    ), "Class config name should be overridden"
    assert (
        config["class_config_1"]["type"] == "tests.test_utils.MyTestClass"
    ), "Class type should not be overridden"
    assert (
        config["class_config_2"]["name"] == "overridden_name2"
    ), "Class config 2 name should be overridden"
    assert (
        config["experiment_dir"] == "affe/experiment"
    ), "Experiment dir should be set to affe/experiment"


def test_override_from_cli():
    with tempfile.NamedTemporaryFile(delete=False, suffix=".json") as temp_file:
        temp_file_path = temp_file.name

        subprocess.run(
            [
                "python",
                script_path,
                "--config",
                config_path,
                "--override",
                "_work_dir=tmp_tmp",
                "--override",
                "class_config_1.name=overridden_name",
                "--override",
                "class_config_1.type=5.",
                "--save_to",
                temp_file_path,
            ],
            check=True,
        )

        # Load the resulting config file
        with open(temp_file_path, "r") as f:
            config = json.load(f)

        assert (
            config["experiment_dir"] == "tmp_tmp/experiment"
        ), "Experiment dir should be overridden"
        assert (
            config["class_config_1"]["name"] == "overridden_name"
        ), "Class config name should be overridden"
        assert (
            config["class_config_1"]["type"] == 5.0
        ), "Config type should be overridden to 5.0"



from konfigurator import class_from_path, get_class_path, instantiate_object_from_config


class MyTestClass:
    def __init__(self, name: str, func=None) -> None:
        self._name = name
        self._func = func

    @property
    def name(self) -> str:
        return self._name


def test_my_test_class():
    class_name = get_class_path(MyTestClass)
    assert class_name == "test_instantiation.MyTestClass"


def test_class_from_path():
    class_name = get_class_path(MyTestClass)
    cls = class_from_path(class_name)
    assert cls is MyTestClass


def test_instantiate_object_from_config():
    def non_pickable():
        yield from range(10)

    config = {
        "type": "test_instantiation.MyTestClass",
        "name": "test_instance",
        "func": non_pickable(),
    }
    instance = instantiate_object_from_config(config)
    assert instance.__class__ is MyTestClass
    assert instance.name == "test_instance"



from .utils import (  # noqa: F401
    FieldReference,
    class_from_path,
    get_class_path,
    instantiate_object_from_config,
    load_config,
)

__all__ = [
    "get_class_path",
    "class_from_path",
    "instantiate_object_from_config",
    "load_config",
    "FieldReference",
]



import argparse
import json
import os

from konfigurator import load_config

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Test script for purepyconf")

    parser.add_argument(
        "--config",
        type=os.path.abspath,
        required=True,
        help="Path to the config file, e.g. 'configs/test_config.yaml'",
    )

    parser.add_argument(
        "--override",
        action="append",
        help="Override config values, e.g. --override key=value",
    )

    parser.add_argument(
        "--save_to",
        type=os.path.abspath,
        help="Path to the file to save the config as a json",
    )

    args = parser.parse_args()

    config = load_config(config_path=args.config, overrides=args.override)
    with open(args.save_to, "w") as f:
        json.dump(config, f, indent=4)
